"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.QueryBuilder = void 0;
const mongodb_1 = require("mongodb");
const errorBuilder_1 = require("../errorBuilder");
const validateValueClass_1 = require("../helpers/validateValueClass");
const metadata_1 = require("../global/metadata");
const type_1 = require("../helpers/type");
const bcryptjs_1 = __importDefault(require("bcryptjs"));
const commons_1 = require("../helpers/commons");
const mongoQueryBuilder_1 = require("./mongoQueryBuilder");
const sqlQueryBuilder_1 = require("./sqlQueryBuilder");
class QueryBuilder {
    getTableName() {
        return ((0, commons_1.convertToPlural)(this.constructor.name));
    }
    getColumns() {
        return (0, metadata_1.getMetadataTable)().getColumns(this.constructor);
    }
    getNameColumns() {
        return this.getColumns().map((column) => column.name);
    }
    getRelations() {
        return (0, metadata_1.getMetadataTable)().getRelation(this.constructor);
    }
    getPrimaryKey() {
        const primaryKey = (0, metadata_1.getMetadataTable)().getPrimaryKey(this.constructor);
        if (!primaryKey) {
            throw Error(`primary key not found`);
        }
        return primaryKey;
    }
    getDB() {
        const table = (0, metadata_1.getMetadataTable)().getTable(this.constructor);
        if (!this.db) {
            if (!this.connection) {
                this.connection = (0, metadata_1.getMetadataDbConnection)();
            }
            if (!this.connection) {
                throw Error(`db connection not found`);
            }
            switch (this.connection.type) {
                case "mongodb":
                    this.db = new mongoQueryBuilder_1.MongoQueryBuilder(this.connection.db, table);
                    break;
                case "postgres":
                case "mysql":
                    this.db = new sqlQueryBuilder_1.SqlQueryBuilder(this.connection.db, table, this.connection.type);
                    break;
                default:
                    throw errorBuilder_1.ErrorBuilder.validator("Invalid database");
            }
        }
        return this.db;
    }
    with(params) {
        this.getDB().with(params, this);
        return this;
    }
    select(columns) {
        this.getDB().select(columns);
        return this;
    }
    findById(id) {
        return __awaiter(this, void 0, void 0, function* () {
            if (typeof id == "string" && this.getPrimaryKey().type == mongodb_1.ObjectId) {
                id = (0, type_1.objectId)(id);
            }
            const data = yield this.findOne({ [this.getPrimaryKey().name]: { $eq: id } });
            return data;
        });
    }
    findOne(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = yield this.find(filter);
            return data.length ? data[0] : null;
        });
    }
    find(filter, options) {
        return __awaiter(this, void 0, void 0, function* () {
            let data = yield this.getDB().find(filter, options);
            this.getDB().reset();
            return data;
        });
    }
    paginate(filter, paginate, options) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getDB().paginate(filter, paginate, options);
        });
    }
    count(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getDB().count(filter);
        });
    }
    createOne(item) {
        return __awaiter(this, void 0, void 0, function* () {
            item = yield this.validateParamBeforeSaveDB(item);
            const dataInsert = yield this.getDB().createOne(item);
            return dataInsert;
        });
    }
    updateMany(filter, item) {
        return __awaiter(this, void 0, void 0, function* () {
            item = yield this.validateParamBeforeSaveDB(item, true);
            const [dataFilter, _dataUpdate] = yield Promise.all([
                this.find(filter),
                this.getDB().updateMany(filter, item)
            ]);
            const primaryKeyName = this.getPrimaryKey().name;
            return this.find({ [primaryKeyName]: { $in: dataFilter.map(e => e[primaryKeyName]) } }, { fromCRUD: true });
        });
    }
    delete(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getDB().deleteOne(filter);
        });
    }
    deleteMany(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getDB().deleteMany(filter);
        });
    }
    validateParamBeforeSaveDB(params, isUpdate = false) {
        return __awaiter(this, void 0, void 0, function* () {
            (0, validateValueClass_1.validationColumnErrors)(this.constructor, params, "", isUpdate);
            yield Promise.all(this.getColumns().map((column) => __awaiter(this, void 0, void 0, function* () {
                const key = column.name;
                if (column.type == mongodb_1.ObjectId && params[key] && typeof params[key] == "string") {
                    params[key] = (0, type_1.objectId)(params[key]);
                }
                if (column.type == String && params[key] && column.isPassword) {
                    params[key] = yield bcryptjs_1.default.hash(params[key], 10);
                }
                if (column.default !== undefined && params[key] === undefined && !isUpdate) {
                    params[key] = column.default;
                }
                return;
            })));
            return params;
        });
    }
}
exports.QueryBuilder = QueryBuilder;
