"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlQueryBuilder = void 0;
const decorators_1 = require("../decorators");
const sequelize_1 = require("sequelize");
const errorBuilder_1 = require("../errorBuilder");
const global_1 = require("../global");
const helpers_1 = require("../helpers");
const multiOps = ["$and", "$or"];
const opMapping = [
    ["$and", sequelize_1.Op.and],
    ["$or", sequelize_1.Op.or],
    ["$in", sequelize_1.Op.in],
    ["$nin", sequelize_1.Op.notIn],
    ["$eq", sequelize_1.Op.eq],
    ["$ne", sequelize_1.Op.ne],
    ["$gt", sequelize_1.Op.gt],
    ["$lt", sequelize_1.Op.lt],
    ["$gte", sequelize_1.Op.gte],
    ["$lte", sequelize_1.Op.lte],
    ["$regex", sequelize_1.Op.regexp],
    ["$like", sequelize_1.Op.like],
    ["$notLike", sequelize_1.Op.notLike],
];
const convertOp = (op) => {
    const map = opMapping.find((p) => p[0] === op);
    if (!map) {
        throw errorBuilder_1.ErrorBuilder.validator("Not supported op: " + op);
    }
    return map[1];
};
class SqlQueryBuilder {
    constructor(db, table, sqlProvider) {
        this.db = db;
        this.table = table;
        this.sqlProvider = sqlProvider;
        this.withRelations = [];
        this.selectColumns = [];
    }
    generateWith(params, model) {
        const withRelations = [];
        const relations = (0, global_1.getMetadataTable)().getTable(model).relations;
        if (Array.isArray(params)) {
            params.forEach(e => {
                const findRelation = relations.find((relation) => relation.fieldName == e.model);
                if (findRelation) {
                    if (e.with) {
                        const modelChild = ((findRelation.type == decorators_1.TypeRelation.HAS_MANY || findRelation.type == decorators_1.TypeRelation.HAS_ONE)
                            ? findRelation.refs.modelChild()
                            : findRelation.refs.modelParent());
                        const withChild = this.generateWith(e.with, modelChild);
                        e.with = withChild;
                    }
                    withRelations.push(Object.assign(Object.assign({}, e), { relation: findRelation }));
                }
                else {
                    throw errorBuilder_1.ErrorBuilder.validator(`with ${e.model.toString()} invalid`);
                }
            });
        }
        else {
            const findRelation = relations.find((relation) => relation.fieldName == params.model);
            if (findRelation) {
                if (params.with) {
                    const modelChild = ((findRelation.type == decorators_1.TypeRelation.HAS_MANY || findRelation.type == decorators_1.TypeRelation.HAS_ONE)
                        ? findRelation.refs.modelChild()
                        : findRelation.refs.modelParent());
                    const withChild = this.generateWith(params.with, modelChild);
                    params.with = withChild;
                }
                withRelations.push(Object.assign(Object.assign({}, params), { relation: findRelation }));
            }
            else {
                throw errorBuilder_1.ErrorBuilder.validator(`with ${params.model.toString()} invalid`);
            }
        }
        return withRelations;
    }
    with(params, model) {
        this.withRelations = this.generateWith(params, model.constructor);
    }
    generateInclude(withRelations, table, isFromCRUD) {
        return withRelations.map(_with => {
            const relation = _with.relation;
            const model = relation.type == decorators_1.TypeRelation.BELONG_TO
                ? relation.refs.modelParent()
                : relation.refs.modelChild();
            const tableInclude = (0, global_1.getMetadataTable)().getTable(model);
            const fieldAllow = tableInclude.columns
                .filter(column => !column.isPassword || !isFromCRUD)
                .map(column => column.name.toString());
            _with.select = _with.select ? _with.select.filter(e => fieldAllow.includes(e.toString())) : fieldAllow;
            const include = {
                association: (0, global_1.getMetadataDbSqlCollections)()[table.name].associations[_with.model.toString()],
                include: _with.with ? this.generateInclude(Array.isArray(_with.with) ? _with.with : [_with.with], tableInclude, isFromCRUD) : [],
                attributes: _with.select,
                as: _with.model.toString(),
                where: _with.filter ? this.getQueryObjectFromQueryConditions(_with.filter) : {},
                required: false
            };
            return include;
        });
    }
    getCollection() {
        return (0, global_1.getMetadataDbSqlCollections)()[this.table.name];
    }
    findByPk(pk) {
        return __awaiter(this, void 0, void 0, function* () {
            const fieldPk = this.table
                .columns
                .find(column => column.isPrimaryKey);
            if (!fieldPk) {
                throw errorBuilder_1.ErrorBuilder.validator(`primary key not found`);
            }
            switch (fieldPk.type) {
                case String:
                    if (typeof pk !== "string") {
                        pk = pk.toString();
                    }
                    break;
                case Number:
                    if (typeof pk === "object") {
                        throw errorBuilder_1.ErrorBuilder.validator(`type ${pk} is not a object`);
                    }
                    pk = +pk;
                    break;
            }
            return yield this.getCollection().findByPk(pk);
        });
    }
    find(filter, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const fieldAllow = this.table
                .columns
                .filter(column => !column.isPassword || !(options === null || options === void 0 ? void 0 : options.fromCRUD))
                .map(column => column.name.toString());
            const selectColumns = this.selectColumns.length
                ? this.selectColumns.filter(e => fieldAllow.includes(e.toString()))
                : fieldAllow;
            const condition = {
                include: this.generateInclude(this.withRelations, this.table),
                where: this.getQueryObjectFromQueryConditions(filter),
                attributes: selectColumns,
            };
            if (options === null || options === void 0 ? void 0 : options.sort) {
                Object.assign(condition, {
                    order: Object.keys(options.sort).reduce((target, item) => {
                        let type = options.sort[item];
                        if (typeof options.sort[item] == "number") {
                            target[item] = options.sort[item] == 1 ? "asc" : "desc";
                        }
                        target.push([
                            item, type.toUpperCase()
                        ]);
                        return target;
                    }, [])
                });
            }
            if (options === null || options === void 0 ? void 0 : options.limit) {
                Object.assign(condition, {
                    limit: options === null || options === void 0 ? void 0 : options.limit
                });
            }
            if (options === null || options === void 0 ? void 0 : options.skip) {
                Object.assign(condition, {
                    offset: options.skip
                });
            }
            return (0, helpers_1.cloneObj)(yield this.getCollection().findAll(condition));
        });
    }
    paginate(filter, paginate, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const fieldAllow = this.table
                .columns
                .filter(column => !column.isPassword || !(options === null || options === void 0 ? void 0 : options.fromCRUD))
                .map(column => column.name.toString());
            const selectColumns = this.selectColumns.length
                ? this.selectColumns.filter(e => fieldAllow.includes(e.toString()))
                : fieldAllow;
            const condition = {
                include: this.generateInclude(this.withRelations, this.table),
                where: this.getQueryObjectFromQueryConditions(filter),
                attributes: selectColumns,
                limit: paginate.limit,
                offset: (paginate.page - 1) * paginate.limit,
            };
            if (options === null || options === void 0 ? void 0 : options.sort) {
                Object.assign(condition, {
                    order: Object.keys(options.sort).reduce((target, item) => {
                        let type = options.sort[item];
                        if (typeof options.sort[item] == "number") {
                            target[item] = options.sort[item] == 1 ? "asc" : "desc";
                        }
                        target.push([
                            item, type.toUpperCase()
                        ]);
                        return target;
                    }, [])
                });
            }
            const result = yield this.getCollection().findAndCountAll(condition);
            return { data: (0, helpers_1.cloneObj)(result.rows), totalCount: result.count };
        });
    }
    count(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            const condition = {
                where: this.getQueryObjectFromQueryConditions(filter),
            };
            return this.getCollection().count(condition);
        });
    }
    createOne(item) {
        return __awaiter(this, void 0, void 0, function* () {
            const fieldPk = this.table
                .columns
                .find(column => column.isPrimaryKey);
            if (!fieldPk) {
                throw errorBuilder_1.ErrorBuilder.validator(`primary key not found`);
            }
            const newDate = new Date().getTime();
            const dataInsert = yield this.getCollection().create(Object.assign(Object.assign({}, item), { createdAt: newDate, updatedAt: newDate }));
            const data = yield this.find({
                [fieldPk.name]: {
                    $eq: dataInsert[fieldPk.name]
                }
            }, { fromCRUD: true });
            return data[0];
        });
    }
    updateMany(filter, item) {
        return __awaiter(this, void 0, void 0, function* () {
            const newDate = new Date().getTime();
            const condition = {
                where: this.getQueryObjectFromQueryConditions(filter),
            };
            return yield this.getCollection().update(Object.assign(Object.assign({}, item), { updatedAt: newDate }), condition);
        });
    }
    deleteOne(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            const records = yield this.find(filter, { limit: 1 });
            if (!records.length) {
                return [];
            }
            const id = records[0]["id"];
            return this.getCollection().destroy({
                where: this.getQueryObjectFromQueryConditions({
                    id: { $eq: id },
                })
            });
        });
    }
    deleteMany(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            const condition = {
                where: this.getQueryObjectFromQueryConditions(filter),
            };
            return this.getCollection().destroy(condition);
        });
    }
    select(selectColumns) {
        this.selectColumns = selectColumns;
    }
    reset() {
        this.withRelations = [];
        this.selectColumns = [];
    }
    mapToSequelizeOperator(op, attribute, value, usingLiteral) {
        switch (op) {
            case "$like":
                switch (this.sqlProvider) {
                    case "postgres":
                        return {
                            [attribute]: {
                                [sequelize_1.Op.iLike]: sequelize_1.Sequelize.literal(`unaccent(${this.db.escape(`%${value}%`)})`),
                            }
                        };
                    case "mysql":
                        return {
                            [attribute]: {
                                [sequelize_1.Op.like]: sequelize_1.Sequelize.literal(`${this.db.escape(`%${value}%`)}`),
                            }
                        };
                }
            default:
                return {
                    [attribute]: {
                        [convertOp(op)]: value,
                    }
                };
        }
    }
    getQueryObjectFromQueryConditions(queries, usingLiteral) {
        if (!queries) {
            return {};
        }
        const keys = Object.keys(queries);
        if (!keys.length) {
            return {};
        }
        const ops = keys.map((key) => {
            if (multiOps.includes(key)) {
                return {
                    [convertOp(key)]: queries[key].map((o) => this.getQueryObjectFromQueryConditions(o))
                };
            }
            else {
                const operator = Object.keys(queries[key])[0];
                const field = key;
                const value = queries[field][operator];
                return {
                    [convertOp("$and")]: this.mapToSequelizeOperator(operator, field, value, usingLiteral),
                };
            }
        });
        return ops.length > 1 ? { [convertOp("$and")]: ops } : ops[0];
    }
}
exports.SqlQueryBuilder = SqlQueryBuilder;
